#ifndef __DALI_TOOLKIT_SCROLL_VIEW_CONSTRAINTS_H__
#define __DALI_TOOLKIT_SCROLL_VIEW_CONSTRAINTS_H__

/*
 * Copyright (c) 2014 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// EXTERNAL INCLUDES

// INTERNAL INCLUDES

namespace Dali
{

struct Vector2;
struct Vector3;
struct Vector4;
class PropertyInput;

namespace Toolkit
{

// Constraints ////////////////////////////////////////////////////////////////////////////////////

/**
 * Useful constraints to apply to a ScrollView.
 */

/**
 * Move Actor constraint.
 *
 * Moves an Actor in accordance to scroll position.
 */
DALI_IMPORT_API Vector3 MoveActorConstraint(const Vector3&    current,
                                           const PropertyInput& scrollPositionProperty);

/**
 * Wrap Actor constraint.
 *
 * Wraps an Actors position in accordance to min/max bounds of domain.
 */
DALI_IMPORT_API Vector3 WrapActorConstraint(const Vector3&    current,
                                            const PropertyInput& actorScaleProperty,
                                            const PropertyInput& actorAnchorPointProperty,
                                            const PropertyInput& actorSizeProperty,
                                            const PropertyInput& scrollPositionMin,
                                            const PropertyInput& scrollPositionMax,
                                            const PropertyInput& scrollWrap);

} // namespace Toolkit

} // namespace Dali

#endif // __DALI_TOOLKIT_SCROLL_VIEW_CONSTRAINTS_H__
